#ifndef _ONE_EURO_FILTER_HPP_
#define _ONE_EURO_FILTER_HPP_

#define AIR_NEON_OPTED

//#if defined | __ANDROID__ || defined __linux__
//#define AIR_NEON_OPTED
//#else
//#define AIR_AVX_OPTED
//#endif

#ifdef AIR_AVX_OPTED
#include <immintrin.h>
#include <xmmintrin.h>
#include <emmintrin.h>
#include <smmintrin.h>
#endif

#ifdef AIR_NEON_OPTED
#include <arm_neon.h>
#endif


class LowPassFilter
{
public:
    int m_num;
    float *m_lastx, *m_alpha, *m_hatxpre, *m_hatx;
    bool initialized;

    LowPassFilter(int num, float *alpha, float initval = 0.f)
    {
        m_num = num;
        //if (alpha <= 0.f || alpha > 1.f)
        //    printf("alpha should be in (0.0., 1.0]");
        m_alpha = alpha;
        // m_alpha = new float[num];
        // for (int i = 0; i < m_num; ++i)
        // {
        //     m_alpha[i] = alpha[i];
        // }
        //lastx = hatxpre = initval;
        m_lastx = new float[num];
        m_hatxpre = new float[num];
        memset(m_lastx, 0, m_num * sizeof(float));
        memset(m_hatxpre, 0, m_num * sizeof(float));
        m_hatx = new float[num];

        initialized = false;
    }
    ~LowPassFilter()
    {
        if (m_lastx)
        {
            delete[] m_lastx;
            m_lastx = NULL;
        }
        if (m_hatxpre)
        {
            delete[] m_hatxpre;
            m_hatxpre = NULL;
        }
        if (m_hatx)
        {
            delete[] m_hatx;
            m_hatx = NULL;
        }
    }
    
    void initialize()
    {
        initialized = false;
        memset(m_lastx, 0, m_num * sizeof(float));
        memset(m_hatxpre, 0, m_num * sizeof(float));
    }

    float *filter(float *x)
    {
        if (initialized)
        {
            // for (int i = 0; i < m_num; ++i)
            // {
            //     //m_hatx[i] = m_alpha[i] * x[i] + (1.f - m_alpha[i]) * m_hatxpre[i];
            //     m_hatx[i] = m_alpha[i] * (x[i] - m_hatxpre[i]) + m_hatxpre[i];
            // }
            int size = m_num;
            float *p_alpha = m_alpha;
            float *p_x = x;
            float *p_hatxpre = m_hatxpre;
            float *p_hatx = m_hatx;
#ifdef AIR_AVX_OPTED
            if (size >= 8)
            {
                int nBlockWidth = 8; //256
                const int cntBlock = size / nBlockWidth;
                int cntRem = size % nBlockWidth;
                size = cntRem;

                __m256 mm_one = _mm256_set1_ps(1.f);
                __m256 mm_alpha, mm_x, mm_hatxpre, mm_res;
                for (int g = 0; g < cntBlock; ++g)
                {
                    mm_alpha = _mm256_loadu_ps(p_alpha);
                    mm_x = _mm256_loadu_ps(p_x);
                    mm_hatxpre = _mm256_loadu_ps(p_hatxpre);

                    mm_res = _mm256_fmadd_ps(_mm256_sub_ps(mm_x, mm_hatxpre), mm_alpha, mm_hatxpre);
                    _mm256_storeu_ps(p_hatx, mm_res);

                    p_alpha += nBlockWidth;
                    p_x += nBlockWidth;
                    p_hatxpre += nBlockWidth;
                    p_hatx += nBlockWidth;
                }

            }
            if (size >= 4)
            {
                int nBlockWidth = 4; //128
                const int cntBlock = size / nBlockWidth;
                int cntRem = size % nBlockWidth;
                size = cntRem;

                __m128 mm_one = _mm_set1_ps(1.f);
                __m128 mm_alpha, mm_x, mm_hatxpre, mm_res;
                for (int g = 0; g < cntBlock; ++g)
                {
                    mm_alpha = _mm_loadu_ps(p_alpha);
                    mm_x = _mm_loadu_ps(p_x);
                    mm_hatxpre = _mm_loadu_ps(p_hatxpre);

                    mm_res = _mm_fmadd_ps(_mm_sub_ps(mm_x, mm_hatxpre), mm_alpha, mm_hatxpre);
                    _mm_storeu_ps(p_hatx, mm_res);

                    p_alpha += nBlockWidth;
                    p_x += nBlockWidth;
                    p_hatxpre += nBlockWidth;
                    p_hatx += nBlockWidth;
                }
            }
#endif
#ifdef AIR_NEON_OPTED

            if (size >= 4)
            {
                int nBlockWidth = 4; //128
                const int cntBlock = size / nBlockWidth;
                int cntRem = size % nBlockWidth;
                size = cntRem;

                float32x4_t mm_alpha, mm_x, mm_res, mm_hatxpre;
                for (int g = 0; g < cntBlock; ++g)
                {
                    mm_alpha = vld1q_f32(p_alpha);
                    mm_x = vld1q_f32(p_x);
                    mm_hatxpre = vld1q_f32(p_hatxpre);

                    mm_res = vaddq_f32(vmulq_f32(vsubq_f32(mm_x, mm_hatxpre), mm_alpha), mm_hatxpre);
                    vst1q_f32(p_hatx, mm_res);

                    p_alpha += nBlockWidth;
                    p_x += nBlockWidth;
                    p_hatxpre += nBlockWidth;
                    p_hatx += nBlockWidth;
                }
            }
#endif
            if (size > 0)
            {
                for (int i = 0; i < size; ++i)
                {
                    //m_hatx[i] = m_alpha[i] * x[i] + (1.f - m_alpha[i]) * m_hatxpre[i];
                    p_hatx[i] = p_alpha[i] * (p_x[i] - p_hatxpre[i]) + p_hatxpre[i];
                }
            }
        }
        else
        {
            memcpy(m_hatx, x, m_num * sizeof(float));
            initialized = true;
        }
        memcpy(m_lastx, x, m_num * sizeof(float));
        memcpy(m_hatxpre, m_hatx, m_num * sizeof(float));
        return m_hatx;
    }

    float *filterWithAlpha(float *x, float *alpha)
    {
        // for (int i = 0; i < m_num; ++i)
        // {
        //     m_alpha[i] = alpha[i];
        // }
        return filter(x);
    }

    bool hasLastRawValue(void)
    {
        return initialized;
    }

    float *lastRawValue(void)
    {
        return m_lastx;
    }
};

// -----------------------------------------------------------------

class OneEuroFilter
{
public:
    int m_num;
    float m_freq;
    float m_beta;
    float m_mincutoff;
    float m_dcutoff;

    float *m_cutoff;
    float *dx;
    float *dxalpha;
    float *xalpha;

    LowPassFilter *xfilter;
    LowPassFilter *dxfilter;

    OneEuroFilter(int num, float freq, float mincutoff = 1.f, float beta_ = 0.f, float dcutoff = 1.f)
    {
        if (num <= 0)
            printf("num should be > 0");
        if (freq <= 0.f)
            printf("freq should be > 0");
        if (mincutoff <= 0.f)
            printf("mincutoff should be > 0");
        if (dcutoff <= 0.f)
            printf("dcutoff should be > 0");

        m_num = num;
        m_freq = freq;
        m_mincutoff = mincutoff;
        m_beta = beta_;
        m_dcutoff = dcutoff;

        m_cutoff = new float[num];//change
        dx = new float[num]; //change
        dxalpha = new float[num]; //change
        xalpha = new float[num]; //change

        alpha(m_mincutoff, xalpha);
        xfilter = new LowPassFilter(num, xalpha);
        alpha(m_dcutoff, dxalpha);
        dxfilter = new LowPassFilter(num, dxalpha);
    }
    
    void initialize()
    {
        xfilter->initialize();
        memset(m_cutoff, 0, m_num * sizeof(float));
        memset(dx, 0, m_num * sizeof(float));
        memset(dxalpha, 0, m_num * sizeof(float));
        memset(xalpha, 0, m_num * sizeof(float));        
    }
    
    ~OneEuroFilter(void)
    {
        if (xfilter)
        {
            delete xfilter;
            xfilter = NULL;
        }
        if (dxfilter)
        {
            delete dxfilter;
            dxfilter = NULL;
        }
        if (m_cutoff)
        {
            delete[] m_cutoff;
            m_cutoff = NULL;
        }
        if (dx)
        {
            delete[] dx;
            dx = NULL;
        }
        if (dxalpha)
        {
            delete[] dxalpha;
            dxalpha = NULL;
        }
        if (xalpha)
        {
            delete[] xalpha;
            xalpha = NULL;
        }
    }


    void filter(float *x)
    {
        // estimate the current variation per second
        // FIXME: 0.0 or value?
        if (xfilter->hasLastRawValue())
        {
            // for (int i = 0; i < m_num; ++i)
            // {
            //     dx[i] = (x[i] - xfilter->lastRawValue()[i]) * m_freq;
            // }
            int size = m_num;
            float *p_x = x;
            float *p_lastx = xfilter->lastRawValue();
            float *p_dx = dx;

#ifdef AIR_AVX_OPTED
            if (size >= 8)
            {
                int nBlockWidth = 8; //256
                const int cntBlock = size / nBlockWidth;
                int cntRem = size % nBlockWidth;
                size = cntRem;

                __m256 mm_freq = _mm256_set1_ps(m_freq);
                __m256 mm_x, mm_lastx, mm_res;
                for (int g = 0; g < cntBlock; ++g)
                {
                    mm_x = _mm256_loadu_ps(p_x);
                    mm_lastx = _mm256_loadu_ps(p_lastx);

                    mm_res = _mm256_mul_ps(_mm256_sub_ps(mm_x, mm_lastx), mm_freq);
                    _mm256_storeu_ps(p_dx, mm_res);

                    p_x += nBlockWidth;
                    p_lastx += nBlockWidth;
                    p_dx += nBlockWidth;
                }
            }
            if (size >= 4)
            {
                int nBlockWidth = 4; //128
                const int cntBlock = size / nBlockWidth;
                int cntRem = size % nBlockWidth;
                size = cntRem;

                __m128 mm_freq = _mm_set1_ps(m_freq);
                __m128 mm_x, mm_lastx, mm_res;
                for (int g = 0; g < cntBlock; ++g)
                {
                    mm_x = _mm_loadu_ps(p_x);
                    mm_lastx = _mm_loadu_ps(p_lastx);

                    mm_res = _mm_mul_ps(_mm_sub_ps(mm_x, mm_lastx), mm_freq);
                    _mm_storeu_ps(p_dx, mm_res);

                    p_x += nBlockWidth;
                    p_lastx += nBlockWidth;
                    p_dx += nBlockWidth;
                }
            }
#endif
#ifdef AIR_NEON_OPTED

            if (size >= 4)
            {
                int nBlockWidth = 4; //128
                const int cntBlock = size / nBlockWidth;
                int cntRem = size % nBlockWidth;
                size = cntRem;

                float32x4_t mm_freq = vdupq_n_f32(m_freq);
                float32x4_t mm_x, mm_lastx, mm_res;
                for (int g = 0; g < cntBlock; ++g)
                {
                    mm_x = vld1q_f32(p_x);
                    mm_lastx = vld1q_f32(p_lastx);

                    mm_res = vmulq_f32(vsubq_f32(mm_x, mm_lastx), mm_freq);
                    vst1q_f32(p_dx, mm_res);

                    p_x += nBlockWidth;
                    p_lastx += nBlockWidth;
                    p_dx += nBlockWidth;
                }
            }
#endif
            if (size > 0)
            {
                for (int i = 0; i < size; ++i)
                {
                    p_dx[i] = (p_x[i] - p_lastx[i]) * m_freq;
                }
            }
        }
        else
        {
            memset(dx, 0, m_num * sizeof(float));
        }

        //alpha(m_dcutoff, dxalpha);
        float *edx = dxfilter->filterWithAlpha(dx, dxalpha);
        // use it to update the cutoff frequency
        for (int i = 0; i < m_num; ++i)
        {
            m_cutoff[i] = m_mincutoff + m_beta * fabs(edx[i]);
        }

        //float m_cutoff = m_mincutoff + m_beta * fabs(edx);
        // filter the given value
        alpha(m_cutoff, xalpha);
        float *res = xfilter->filterWithAlpha(x, xalpha);
        memcpy(x, res, m_num * sizeof(float));
    }

    void alpha(float *cutoff, float *resalpha)
    {
        //float tmp = 2.f * 3.141593f * cutoff;
        //return tmp / (tmp + m_freq);
        // for (int i = 0; i < m_num; ++i)
        // {
        //     float tmp = 2.f * 3.141593f * cutoff[i];
        //     resalpha[i] = tmp / (tmp + m_freq);
        // }
        int size = m_num;
        float *p_cutoff = cutoff;
        float *p_resalpha = resalpha;

#ifdef AIR_AVX_OPTED
        if (size >= 8)
        {
            int nBlockWidth = 8; //256
            const int cntBlock = size / nBlockWidth;
            int cntRem = size % nBlockWidth;
            size = cntRem;

            __m256 mm_freq = _mm256_set1_ps(m_freq);
            __m256 mm_2pi = _mm256_set1_ps(2.f * 3.141593f);
            __m256 mm_cutoff, mm_alpha, mm_tmp;
            for (int g = 0; g < cntBlock; ++g)
            {
                mm_cutoff = _mm256_loadu_ps(p_cutoff);
                mm_tmp = _mm256_mul_ps(mm_2pi, mm_cutoff);

                mm_alpha = _mm256_div_ps(mm_tmp, _mm256_add_ps(mm_tmp, mm_freq));
                _mm256_storeu_ps(p_resalpha, mm_alpha);

                p_cutoff += nBlockWidth;
                p_resalpha += nBlockWidth;
            }
        }
        if (size >= 4)
        {
            int nBlockWidth = 4; //128
            const int cntBlock = size / nBlockWidth;
            int cntRem = size % nBlockWidth;
            size = cntRem;

            __m128 mm_freq = _mm_set1_ps(m_freq);
            __m128 mm_2pi = _mm_set1_ps(2.f * 3.141593f);
            __m128 mm_cutoff, mm_alpha, mm_tmp;
            for (int g = 0; g < cntBlock; ++g)
            {
                mm_cutoff = _mm_loadu_ps(p_cutoff);
                mm_tmp = _mm_mul_ps(mm_2pi, mm_cutoff);

                mm_alpha = _mm_div_ps(mm_tmp, _mm_add_ps(mm_tmp, mm_freq));
                _mm_storeu_ps(p_resalpha, mm_alpha);

                p_cutoff += nBlockWidth;
                p_resalpha += nBlockWidth;
            }
        }
#endif
#ifdef AIR_NEON_OPTED

        if (size >= 4)
        {
            int nBlockWidth = 4; //128
            const int cntBlock = size / nBlockWidth;
            int cntRem = size % nBlockWidth;
            size = cntRem;

            float32x4_t mm_freq = vdupq_n_f32(m_freq);
            float32x4_t mm_2pi = vdupq_n_f32(2.f * 3.141593f);
            float32x4_t mm_cutoff, mm_alpha, mm_tmpa, mm_tmpb, recip;
            for (int g = 0; g < cntBlock; ++g)
            {
                mm_cutoff = vld1q_f32(p_cutoff);
                mm_tmpa = vmulq_f32(mm_2pi, mm_cutoff);

                //mm_alpha = vdivq_f32(mm_tmp, vaddq_f32(mm_tmpa, mm_freq));

                // compute a/b, neon has not direct function for a/b
                // 1. compute 1/b
                mm_tmpb = vaddq_f32(mm_tmpa, mm_freq);
                recip = vrecpeq_f32(mm_tmpb); /*此时能达到千分之一左右的精度，如1.0的倒数为0.998047*/
                recip = vmulq_f32 (vrecpsq_f32 (mm_tmpb, recip), recip); /*执行后能达到百万分之一左右的精度，如1.0的倒数为0.999996*/
                recip = vmulq_f32 (vrecpsq_f32 (mm_tmpb, recip), recip); /*再次执行后能基本能达到完全精度，如1.0的倒数为1.000000*/
                // 2. compute a/b = a*(1/b)
                mm_alpha = vmulq_f32(mm_tmpa, recip);

                vst1q_f32(p_resalpha, mm_alpha);
                p_cutoff += nBlockWidth;
                p_resalpha += nBlockWidth;
            }
        }
#endif
        if (size > 0)
        {
            for (int i = 0; i < size; ++i)
            {
                float tmp = 2.f * 3.141593f * p_cutoff[i];
                p_resalpha[i] = tmp / (tmp + m_freq);
            }
        }
    }
    void alpha(float cutoff, float *resalpha)
    {
        // float tmp = 2.f * 3.141593f * cutoff;
        // float palpha = tmp / (tmp + m_freq);
        // for (int i = 0; i < m_num; ++i)
        // {
        //     resalpha[i] = palpha;
        // }
        int size = m_num;
        float *p_resalpha = resalpha;
#ifdef AIR_AVX_OPTED
        if (size >= 8)
        {
            int nBlockWidth = 8; //256
            const int cntBlock = size / nBlockWidth;
            int cntRem = size % nBlockWidth;
            size = cntRem;

            float tmp = 2.f * 3.141593f * cutoff;
            float palpha = tmp / (tmp + m_freq);
            __m256 mm_alpha = _mm256_set1_ps(palpha);
            for (int g = 0; g < cntBlock; ++g)
            {
                _mm256_storeu_ps(p_resalpha, mm_alpha);
                p_resalpha += nBlockWidth;
            }
        }
        if (size >= 4)
        {
            int nBlockWidth = 4; //128
            const int cntBlock = size / nBlockWidth;
            int cntRem = size % nBlockWidth;
            size = cntRem;

            float tmp = 2.f * 3.141593f * cutoff;
            float palpha = tmp / (tmp + m_freq);
            __m128 mm_alpha = _mm_set1_ps(palpha);
            for (int g = 0; g < cntBlock; ++g)
            {
                _mm_storeu_ps(p_resalpha, mm_alpha);
                p_resalpha += nBlockWidth;
            }
        }
#endif
#ifdef AIR_NEON_OPTED
        if (size >= 4)
        {
            int nBlockWidth = 4; //128
            const int cntBlock = size / nBlockWidth;
            int cntRem = size % nBlockWidth;
            size = cntRem;

            float tmp = 2.f * 3.141593f * cutoff;
            float palpha = tmp / (tmp + m_freq);
            float32x4_t mm_alpha = vdupq_n_f32(palpha);
            for (int g = 0; g < cntBlock; ++g)
            {
                vst1q_f32(p_resalpha, mm_alpha);
                p_resalpha += nBlockWidth;
            }
        }
#endif
        if (size > 0)
        {
            float tmp = 2.f * 3.141593f * cutoff;
            float palpha = tmp / (tmp + m_freq);
            for (int i = 0; i < size; ++i)
            {
                p_resalpha[i] = palpha;
            }
        }
    }
};

#endif

