#ifndef _BILINER_RESIZE_HPP_
#define _BILINER_RESIZE_HPP_
#include <math.h>
#include "CL/cl.h"
#include "include/cltool.h"
#include <string.h>

class BilinerResize
{
public:
    BilinerResize(unsigned char *src, int src_width, int src_height,
                    unsigned char *dst, int dst_width, int dst_height);
    ~BilinerResize();
    void run(unsigned char *src, int src_width, int src_height,
            unsigned char *dst, int dst_width, int dst_height);

private:
    int pre_src_width_;
    int pre_src_height_;
    int pre_dst_width_;
    int pre_dst_height_;

    float *x_scale_;
    float *y_scale_;
    unsigned int *x_left_crood_;
    unsigned int *y_left_crood_;
    
    unsigned char *pre_src_;
    unsigned char *pre_dst_;
    //cl param
    cl_context context_;
    cl_command_queue command_queue_;
    cl_program program_;
    cl_kernel kernel_;

    cl_mem src_buffer_;
    cl_mem y_scale_buffer_;
    cl_mem x_scale_buffer_;
    cl_mem x_left_crood_buffer_;
    cl_mem y_left_crood_buffer_;
    cl_mem dst_buffer_;

    void *x_left_crood_ptr_;
    void *y_left_crood_ptr_;
    void *y_scale_ptr_;
    void *x_scale_ptr_;
    void *src_ptr_;
    void *dst_ptr_;

};

BilinerResize::BilinerResize(unsigned char *src, int src_width, int src_height,
                            unsigned char *dst, int dst_width, int dst_height)
{
    pre_src_width_ = src_width;
    pre_src_height_ = src_height;
    pre_dst_width_ = dst_width;
    pre_dst_height_ = dst_height;

    pre_src_ = src;
    pre_dst_ = dst;

    x_scale_ = NULL;
    y_scale_ = NULL;
    x_left_crood_ = NULL;
    y_left_crood_ = NULL;
    
    if(!x_scale_ && !y_scale_ && !x_left_crood_ && !y_left_crood_)
    {
        x_scale_ = (float *)malloc(1280 * sizeof(float));
        y_scale_ = (float *)malloc(1280 * sizeof(float));
        x_left_crood_ = (unsigned int *)malloc(1280 * sizeof(int));
        y_left_crood_ = (unsigned int *)malloc(1280 * sizeof(int));
    }
    

    //opencl init
    cl_int    status;
    /**Step 1: Getting platforms and choose an available one(first).*/
    cl_platform_id platform;
    getPlatform(platform);
    printf("testing\n");
    /**Step 2:Query the platform and choose the first GPU device if has one.*/
    cl_device_id *devices=getCl_device_id(platform);
    
    //get device info
    getCl_device_info(devices[0]);

    /**Step 3: Create context.*/
    context_ = clCreateContext(NULL,1, devices,NULL,NULL,NULL);

    /**Step 4: Creating command queue associate with the context.*/
    command_queue_ = clCreateCommandQueue(context_, devices[0], 0, NULL);

    /**Step 5: Create program object */
    const char *filename = "kernel.cl";

    string sourceStr;
    status = convertToString(filename, sourceStr);
    const char *source = sourceStr.c_str();
    size_t sourceSize[] = {strlen(sourceStr.c_str())};
    program_ = clCreateProgramWithSource(context_, 1, &source, sourceSize, NULL);

    /**Step 6: Build program. */
    status=clBuildProgram(program_, 1, devices, NULL, NULL, NULL);
    if(CL_SUCCESS != status)
    {
        printf("build program failed: %d\n", status);
        char param_value[4096];
        size_t param_value_size = 4096;
        status = clGetProgramBuildInfo(program_, devices[0], CL_PROGRAM_BUILD_LOG, param_value_size, param_value, &param_value_size);
        if(CL_SUCCESS != status)
        {
            printf("clGetProgramBuildInfo failed: %d\n", status);
        }
        printf("clGetProgramBuildInfo: %d\n", param_value_size);
        for(int i = 0; i < param_value_size; i++)
        {
            printf("%c", param_value[i]);
        }
        printf("\n");
    }
    kernel_ = clCreateKernel(program_,"biliner", NULL);

    src_buffer_ = clCreateBuffer(context_, CL_MEM_READ_ONLY|CL_MEM_USE_HOST_PTR, src_width * src_height,(void *)src, NULL);
    y_scale_buffer_ = clCreateBuffer(context_, CL_MEM_READ_ONLY|CL_MEM_USE_HOST_PTR, 1280 * sizeof(float),(void *)y_scale_, NULL);
    x_scale_buffer_ = clCreateBuffer(context_, CL_MEM_READ_ONLY|CL_MEM_USE_HOST_PTR, 1280 * sizeof(float),(void *)x_scale_, NULL);
    x_left_crood_buffer_ = clCreateBuffer(context_, CL_MEM_READ_ONLY|CL_MEM_USE_HOST_PTR, 1280 * sizeof(int),(void *)x_left_crood_, NULL);
    y_left_crood_buffer_ = clCreateBuffer(context_, CL_MEM_READ_ONLY|CL_MEM_USE_HOST_PTR, 1280 * sizeof(int),(void *)y_left_crood_, NULL);

    y_scale_ptr_ = clEnqueueMapBuffer(command_queue_, y_scale_buffer_, CL_TRUE, CL_MAP_WRITE, 0, dst_height * sizeof(float), 0, NULL, NULL, NULL);
    x_scale_ptr_ = clEnqueueMapBuffer(command_queue_, x_scale_buffer_, CL_TRUE, CL_MAP_WRITE, 0, dst_width * sizeof(float), 0, NULL, NULL, NULL);
    x_left_crood_ptr_ = clEnqueueMapBuffer(command_queue_, x_left_crood_buffer_, CL_TRUE, CL_MAP_WRITE, 0, dst_width * sizeof(int), 0, NULL, NULL, NULL);
    y_left_crood_ptr_ = clEnqueueMapBuffer(command_queue_, y_left_crood_buffer_, CL_TRUE, CL_MAP_WRITE, 0, dst_height * sizeof(int), 0, NULL, NULL, NULL);


    float scale_x = (float)(src_width - 1) / (dst_width - 1);
    float scale_y = (float)(src_height - 1) / (dst_height - 1);

    for(int w = 0; w<dst_width; w++)
    {
        if(w == 0)
        {
            x_left_crood_[w] = 0;
            x_scale_[w] = 0;
        } else if(w == dst_width - 1)
        {
            x_left_crood_[w] = src_width - 2;
            x_scale_[w] = 1;
        } else
        {
            x_left_crood_[w] = floorf(w * scale_x);
            // x_scale_[w] = (w * scale_x - x_left_crood_[w]) * 255;
            x_scale_[w] = (w * scale_x - x_left_crood_[w]);

        }
    }
    for(int h = 0; h<dst_height; h++)
    {
        if(h == 0)
        {
            y_left_crood_[h] = 0;
            y_scale_[h] = 0;
        }  else if(h == dst_height - 1)
        {
            y_left_crood_[h] = src_height - 2;
            y_scale_[h] = 1;
        } else 
        {
            y_left_crood_[h] = floorf(h * scale_y);
            // y_scale_[h] = (h * scale_y - y_left_crood_[h]) * 255;
            y_scale_[h] = (h * scale_y - y_left_crood_[h]);
        }
    }
    clEnqueueUnmapMemObject(command_queue_, y_scale_buffer_, y_scale_ptr_, 0, NULL, NULL);
    clEnqueueUnmapMemObject(command_queue_, x_scale_buffer_, x_scale_ptr_, 0, NULL, NULL);
    clEnqueueUnmapMemObject(command_queue_, x_left_crood_buffer_, x_left_crood_ptr_, 0, NULL, NULL);
    clEnqueueUnmapMemObject(command_queue_, y_left_crood_buffer_, y_left_crood_ptr_, 0, NULL, NULL);

    dst_buffer_ = clCreateBuffer(context_, CL_MEM_WRITE_ONLY|CL_MEM_USE_HOST_PTR, dst_width * dst_height, (void *)dst, NULL);

    status = clSetKernelArg(kernel_, 0, sizeof(cl_mem), (void *)&src_buffer_);
    status = clSetKernelArg(kernel_, 1, sizeof(cl_mem), (void *)&dst_buffer_);
    status = clSetKernelArg(kernel_, 2, sizeof(cl_mem), (void *)&y_scale_buffer_);
    status = clSetKernelArg(kernel_, 3, sizeof(cl_mem), (void *)&x_scale_buffer_);
    status = clSetKernelArg(kernel_, 4, sizeof(cl_mem), (void *)&x_left_crood_buffer_);
    status = clSetKernelArg(kernel_, 5, sizeof(cl_mem), (void *)&y_left_crood_buffer_);

    // status = clSetKernelArg(kernel_, 6, sizeof(int), (void *)&src_width);
    // status = clSetKernelArg(kernel_, 7, sizeof(int), (void *)&src_height);
    // status = clSetKernelArg(kernel_, 8, sizeof(int), (void *)&dst_width);
    // status = clSetKernelArg(kernel_, 9, sizeof(int), (void *)&dst_height);
}

BilinerResize::~BilinerResize()
{
    if(x_scale_)
        free(x_scale_);
    if(y_scale_)
        free(y_scale_);
    if(x_left_crood_)
        free(x_left_crood_);
    if(y_left_crood_)
        free(y_left_crood_);

    clReleaseMemObject(src_buffer_);
    clReleaseMemObject(y_scale_buffer_);
    clReleaseMemObject(x_scale_buffer_);
    clReleaseMemObject(x_left_crood_buffer_);
    clReleaseMemObject(y_left_crood_buffer_);
    clReleaseMemObject(dst_buffer_);
}

void BilinerResize::run(unsigned char *src, int src_width, int src_height,
                        unsigned char *dst, int dst_width, int dst_height)
{
#if 0
    //pre process
    if(dst_width > 1280 || dst_height > 1280)
    {
        printf("invilid dst_width or dst_height, dst_width and dst_height should <= 1280\n");
    }
    if(dst_width != pre_dst_width_ || dst_height != pre_dst_height_ || src_width != pre_src_width_ || src_height != pre_src_height_)
    {
        y_scale_ptr_ = clEnqueueMapBuffer(command_queue_, y_scale_buffer_, CL_TRUE, CL_MAP_WRITE, 0, dst_height * sizeof(float), 0, NULL, NULL, NULL);
        x_scale_ptr_ = clEnqueueMapBuffer(command_queue_, x_scale_buffer_, CL_TRUE, CL_MAP_WRITE, 0, dst_width * sizeof(float), 0, NULL, NULL, NULL);
        x_left_crood_ptr_ = clEnqueueMapBuffer(command_queue_, x_left_crood_buffer_, CL_TRUE, CL_MAP_WRITE, 0, dst_width * sizeof(int), 0, NULL, NULL, NULL);
        y_left_crood_ptr_ = clEnqueueMapBuffer(command_queue_, y_left_crood_buffer_, CL_TRUE, CL_MAP_WRITE, 0, dst_height * sizeof(int), 0, NULL, NULL, NULL);

        float scale_x = (float)(src_width - 1) / (dst_width - 1);
        float scale_y = (float)(src_height - 1) / (dst_height - 1);

        for(int w = 0; w<dst_width; w++)
        {
            if(w == 0)
            {
                x_left_crood_[w] = 0;
                x_scale_[w] = 0;
            } else if(w == dst_width - 1)
            {
                x_left_crood_[w] = src_width - 2;
                x_scale_[w] = 1;
            } else
            {
                x_left_crood_[w] = floorf(w * scale_x);
                // x_scale_[w] = (w * scale_x - x_left_crood_[w]) * 255;
                x_scale_[w] = (w * scale_x - x_left_crood_[w]);

            }
        }
        for(int h = 0; h<dst_height; h++)
        {
            if(h == 0)
            {
                y_left_crood_[h] = 0;
                y_scale_[h] = 0;
            }  else if(h == dst_height - 1)
            {
                y_left_crood_[h] = src_height - 2;
                y_scale_[h] = 1;
            } else 
            {
                y_left_crood_[h] = floorf(h * scale_y);
                // y_scale_[h] = (h * scale_y - y_left_crood_[h]) * 255;
                y_scale_[h] = (h * scale_y - y_left_crood_[h]);
            }
        }

        clEnqueueUnmapMemObject(command_queue_, y_scale_buffer_, y_scale_ptr_, 0, NULL, NULL);
        clEnqueueUnmapMemObject(command_queue_, x_scale_buffer_, x_scale_ptr_, 0, NULL, NULL);
        clEnqueueUnmapMemObject(command_queue_, x_left_crood_buffer_, x_left_crood_ptr_, 0, NULL, NULL);
        clEnqueueUnmapMemObject(command_queue_, y_left_crood_buffer_, y_left_crood_ptr_, 0, NULL, NULL);

        //
        clReleaseMemObject(dst_buffer_);
        dst_buffer_ = clCreateBuffer(context_, CL_MEM_WRITE_ONLY|CL_MEM_USE_HOST_PTR, dst_width * dst_height, (void *)dst, NULL);

        clReleaseMemObject(src_buffer_);
        src_buffer_ = clCreateBuffer(context_, CL_MEM_WRITE_ONLY|CL_MEM_USE_HOST_PTR, src_width * src_height, (void *)src, NULL);
    }
    if(pre_dst_ != dst)
    {
        clReleaseMemObject(dst_buffer_);
        dst_buffer_ = clCreateBuffer(context_, CL_MEM_WRITE_ONLY|CL_MEM_USE_HOST_PTR, dst_width * dst_height, (void *)dst, NULL);
    }
    if(pre_src_ != src)
    {
        clReleaseMemObject(src_buffer_);
        src_buffer_ = clCreateBuffer(context_, CL_MEM_WRITE_ONLY|CL_MEM_USE_HOST_PTR, src_width * src_height, (void *)src, NULL);
    }

    src_ptr_ = clEnqueueMapBuffer(command_queue_, src_buffer_, CL_TRUE, CL_MAP_WRITE, 0, src_width * src_height, 0, NULL, NULL, NULL);

    if(dst_ptr_)
    {
        clEnqueueUnmapMemObject(command_queue_, dst_buffer_, dst_ptr_, 0, NULL, NULL);
    }
    pre_src_width_ = src_width;
    pre_src_height_ = src_height;
    pre_dst_width_ = dst_width;
    pre_dst_height_ = dst_height;
#endif

    // clEnqueueUnmapMemObject(command_queue_, y_scale_buffer_, y_scale_ptr_, 0, NULL, NULL);
    // clEnqueueUnmapMemObject(command_queue_, x_scale_buffer_, x_scale_ptr_, 0, NULL, NULL);
    // clEnqueueUnmapMemObject(command_queue_, x_left_crood_buffer_, x_left_crood_ptr_, 0, NULL, NULL);
    // clEnqueueUnmapMemObject(command_queue_, y_left_crood_buffer_, y_left_crood_ptr_, 0, NULL, NULL);
    // if(dst_ptr_)
    // {
    //     clEnqueueUnmapMemObject(command_queue_, dst_buffer_, dst_ptr_, 0, NULL, NULL);
    // }
    // clEnqueueUnmapMemObject(command_queue_, src_buffer_, src_ptr_, 0, NULL, NULL);
    src_ptr_ = clEnqueueMapBuffer(command_queue_, src_buffer_, CL_TRUE, CL_MAP_WRITE, 0, src_width * src_height, 0, NULL, NULL, NULL);
    //opencl run
    cl_int status;
    status = clSetKernelArg(kernel_, 6, sizeof(int), (void *)&src_width);
    status = clSetKernelArg(kernel_, 7, sizeof(int), (void *)&src_height);
    status = clSetKernelArg(kernel_, 8, sizeof(int), (void *)&dst_width);
    status = clSetKernelArg(kernel_, 9, sizeof(int), (void *)&dst_height);

    size_t global_work_size[2];
    global_work_size[0] = dst_width;
    global_work_size[1] = dst_height;

    cl_event enentPoint;
    status = clEnqueueNDRangeKernel(command_queue_, kernel_, 2, NULL, global_work_size, NULL, 0, NULL, &enentPoint);
    if(CL_SUCCESS != status)
    {
        printf("clEnqueueNDRangeKernel failed: %d\n", status);
    }
    clFinish(command_queue_);
    dst_ptr_ = clEnqueueMapBuffer(command_queue_, dst_buffer_, CL_TRUE, CL_MAP_READ, 0, dst_width*dst_height, 0, NULL, NULL, NULL);

    clEnqueueUnmapMemObject(command_queue_, src_buffer_, src_ptr_, 0, NULL, NULL);
    // clWaitForEvents(1,&enentPoint); ///wait
    // clReleaseEvent(enentPoint);

    // /**Step 11: Read the cout put back to host memory.*/
    // status = clEnqueueReadBuffer(command_queue_, dst_buffer_, CL_TRUE, 0, dst_width * dst_height, dst, 0, NULL, NULL);
}
#endif
